from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib
import hmac

import tiros.util as util


def sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def get_signature_key(key, date, region, service):
    k_date = sign(('AWS4' + key).encode('utf-8'), date)
    k_region = sign(k_date, region)
    k_service = sign(k_region, service)
    return sign(k_service, 'aws4_request')


class Signer:
    """
    A record with sufficient information to create a sigv4 signature and
    authorization header for an HTTP request.

    See: http://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
    """
    def __init__(self, access_key, amz_date, body, date_stamp, host, method,
                 region, route, secret_key):
        assert date_stamp
        assert region
        # Host must be lower case (for authentication)
        host = host.lower()
        algorithm = 'AWS4-HMAC-SHA256'
        canonical_querystring = ''
        service = 'tiros'
        signed_headers = 'content-type;host;x-amz-date'
        canonical_headers = (
            'content-type:{}\nhost:{}\nx-amz-date:{}\n'.format(
                util.CONTENT_TYPE, host, amz_date))
        payload_hash = hashlib.sha256(body.encode()).hexdigest()
        canonical_request = '\n'.join(
            [method, route, canonical_querystring, canonical_headers,
             signed_headers, payload_hash])
        credential_scope = '/'.join(
            [date_stamp, region, service, 'aws4_request'])
        string_to_sign = '\n'.join(
            [algorithm, amz_date, credential_scope,
             hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()])
        signing_key = get_signature_key(
            secret_key, date_stamp, region, service)
        credential = ''.join([access_key, '/', credential_scope])
        self.signature = hmac.new(
            signing_key, string_to_sign.encode('utf-8'),
            hashlib.sha256).hexdigest()
        # NB: It's important to quote the auth header, since it has characters
        # not generally legal for HTTP headers like '/'
        self.authorization_header = ''.join([
            algorithm,
            ' Credential=', util.quote(credential),
            ', SignedHeaders=', util.quote(signed_headers),
            ', Signature=', self.signature
        ])


class AuthSession:
    """
    A record of the components of an AWS profile.  It's essentially the
    same record as a boto.Session, so there are convenience methods
    to change back and forth.  It also supports the read syntax
    profile-name/region.
    """
    def __init__(self, session):
        assert session.region_name, 'session needs a region for auth'
        self.session = session

    def access_key(self):
        return self.session.get_credentials().access_key

    def secret_key(self):
        return self.session.get_credentials().secret_key

    def token(self):
        return self.session.get_credentials().token

    def region(self):
        return self.session.region_name

    def _signer(self, amz_date, body, date_stamp, host, method, route):
        return Signer(access_key=self.access_key(),
                      amz_date=amz_date,
                      body=body,
                      date_stamp=date_stamp,
                      host=host,
                      method=method,
                      region=self.region(),
                      route=route,
                      secret_key=self.secret_key())

    def auth_header(self, amz_date, body, date_stamp, host, method, route):
        return self._signer(amz_date, body, date_stamp, host, method, route).authorization_header

    def request_key(self, amz_date, body, date_stamp, host, method, route):
        obj = {
            'accessKeyId': self.access_key(),
            'amzDate': amz_date,
            'body': body,
            'host': host,
            'method': method,
            'region': self.region(),
            'route': route,
            'signature': self._signer(amz_date, body, date_stamp, host, method, route).signature
        }
        # Python encodes None in JSON as "None", not null
        if self.token():
            obj['sessionToken'] = self.token()
        return obj

    def snapshot_key(self, amz_date, date_stamp, host):
        return self.request_key(amz_date, '', date_stamp, host, util.METHOD,
                                'snapshot')



from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import json
import sys

import tiros.fetch_ as fetch_
import tiros.server as server
import tiros.util as util

from tiros.util import eprint, pretty, vprint


def format_response(response, json_errors):
    try:
        j = json.loads(response.text)
    # The exception thrown on json parse failure in python3 is
    # json.decoder.JSONDecodeError, which does not exist in python2.
    # Fortunately, what python2 throws in that case is ValueError,
    # which is a superclass of json.decoder.JSONDecodeError.
    except ValueError:
        j = {'error': response.text}
    if 'error' in j and not json_errors:
        return j['error']['message']
    else:
        return pretty(j)


def snapshot_session(name):
    if '/' not in name:
        session = util.new_session(profile_name=name)
    else:
        [name, region] = name.split('/')
        session = util.new_session(profile_name=name, region_name=region)
    return session


# XXX: Remove this.  The code that calls Tiros on a Palisade snapshot
# should be the one doing the json surgery here.
def autodetect_raw_snapshot(s):
    # For Palisade snapshots, the raw snapshot holds the actual snapshot
    # in the db field
    if 'db' in s:
        return s.get('db')
    else:
        return s


general_args = argparse.ArgumentParser(add_help=False)

general_args.add_argument(
    '--profile',
    help='The AWS profile to use for signing requests'
)

general_args.add_argument(
    '--region',
    help='The AWS region to use for signing requests'
)

general_args.add_argument(
    '--verbose',
    '-v',
    action='store_true',
    help='Be chatty'
)

server_args = argparse.ArgumentParser(add_help=False)

server_args.add_argument(
    '--dev',
    action='store_true',
    help='Use the dev server'
)

server_args.add_argument(
    '--host',
    help='''
    Tiros host.  If the port is nonstandard, include it here,
    e.g. http://localhost:9000
    '''
)

server_args.add_argument(
    '--json',
    action='store_true',
    help='Force json output, even for errors'
)

server_args.add_argument(
    '--no-ssl',
    action='store_true',
    help="Don't use SSL"
)

server_args.add_argument(
    '--raw-snapshot-file',
    '-r',
    default=[],
    action='append',
    help='File containing the raw JSON snapshot'
)

server_args.add_argument(
    '--snapshot-file',
    '-s',
    default=[],
    action='append',
    help='File containing the raw JSON snapshot'
)

server_args.add_argument(
    '--snapshot-profile',
    '-n',
    action='append',
    help="""
    An AWS profile name or profile/region pair.
    (e.g. dev, dev/us-east-1, prod/us-west-2).
    Multiple profiles are supported.  If no profile is
    provided the default is used. If no region
    is specified for a profile, the default region for the
    profile is used.
    """
)

server_args.add_argument(
    '--unsafe-ignore-classic',
    action='store_true',
    help="""
    Silently drop any VPC Classic resources we discover.  You can not
    trust Tiros results when this flag is set!  If ClassicLink is used
    in a VPC, Tiros has no information about the linked Classic instances.
    """
)

parser = argparse.ArgumentParser(
    description='Tiros, version ' + util.TIROS_VERSION)

subparsers = parser.add_subparsers(title='subcommands', dest='cmd')


def main():
    args = parser.parse_args()
    if args.verbose:
        util.VERBOSE = True
    if not args.cmd:
        print("No subcommand specified")
        parser.print_help()
        sys.exit(1)
    args.func(args)


# ---------- fetch -----------


def top_fetch(args):
    session = util.new_session(profile_name=args.profile,
                               region_name=args.region)
    if args.region:
        session = util.change_session_region(session, args.region)
    print(util.pretty(fetch_.fetch(session)))

fetch_parser = subparsers.add_parser(
    'fetch',
    parents=[general_args],
    help='Fetch a tiros snapshot using local credentials'
)

fetch_parser.set_defaults(func=top_fetch)


# ---------- query ----------


query_parser = subparsers.add_parser(
    'query',
    parents=[general_args, server_args],
    help='Query a network using TQL'
)


def query(args):
    session = util.new_session(profile_name=args.profile,
                               region_name=args.region)
    ssl = not args.no_ssl
    host = args.host or (server.DEV_HOST if args.dev else server.PROD_HOST)
    snapshot_sessions = [snapshot_session(p) for p in (args.snapshot_profile or [])]
    snapshots = [json.loads(util.file_contents(f))
                 for f in (args.snapshot_file or [])]
    raw_snapshots = [autodetect_raw_snapshot(json.loads(util.file_contents(f)))
                     for f in (args.raw_snapshot_file or [])]
    if not snapshot_sessions and not snapshots and not raw_snapshots:
        if not session.region_name:
            eprint('No snapshots given and the signing session has no region')
        snapshot_sessions = [session]
    if args.inline and args.queries_file:
        eprint("Can't specify both --inline and --queries_file")
    queries = None
    if args.inline:
        queries = args.inline
    elif args.queries_file:
        queries = util.file_contents(args.queries_file)
    else:
        eprint('You must specify --inline or --queries_file')
    if args.relations_file:
        user_relations = util.file_contents(args.relations_file)
    else:
        user_relations = None
    response = server.query(
        signing_session=session,
        queries=queries,
        backend=args.backend,
        host=host,
        raw_snapshots=raw_snapshots,
        snapshot_sessions=snapshot_sessions,
        snapshots=snapshots,
        ssl=ssl,
        transforms=args.transform,
        unsafe_ignore_classic=args.unsafe_ignore_classic,
        user_relations=user_relations)
    vprint('Status code: ' + str(response.status_code))
    print(format_response(response, args.json))
    if response.status_code != 200:
        sys.exit(1)
    sys.exit(0)

query_parser.set_defaults(func=query)

query_parser.add_argument(
    '--backend',
    '-b',
    help='Datalog backend'
)

query_parser.add_argument(
    '--inline',
    '-i',
    help='Inline query'
)

query_parser.add_argument(
    '--queries-file',
    '-f',
    default=None,
    help='File containing the JSON Tiros queries'
)

query_parser.add_argument(
    '--relations-file',
    '-l',
    default=None,
    help='User relations file'
)

query_parser.add_argument(
    '--transform',
    '-x',
    action='append',
    default=[],
    help='Apply source transforms. Available: magic-sets'
)


# ---------- snapshot ----------


def snapshot(args):
    session = util.new_session(profile_name=args.profile,
                               region_name=args.region)
    ssl = not args.no_ssl
    host = args.host or (server.DEV_HOST if args.dev else server.PROD_HOST)
    snapshot_sessions = [snapshot_session(p) for p in (args.snapshot_profile or [])]
    snapshots = [json.loads(util.file_contents(f))
                 for f in (args.snapshot_file or [])]
    raw_snapshots = [autodetect_raw_snapshot(json.loads(util.file_contents(f)))
                     for f in (args.raw_snapshot_file or [])]
    if not snapshot_sessions and not snapshots and not raw_snapshots:
        snapshot_sessions = [session]
    response = server.snapshot(
        host=host,
        raw_snapshots=raw_snapshots,
        signing_session=session,
        snapshot_sessions=snapshot_sessions,
        snapshots=snapshots,
        ssl=ssl,
        unsafe_ignore_classic=args.unsafe_ignore_classic,
    )
    vprint('Status code: ' + str(response.status_code))
    print(format_response(response, args.json))
    if response.status_code != 200:
        sys.exit(1)
    sys.exit(0)

snapshot_parser = subparsers.add_parser(
    'snapshot',
    parents=[general_args, server_args],
    help="Get a network snapshot using the Tiros service's credentials"
)

snapshot_parser.set_defaults(func=snapshot)


if __name__ == '__main__':
    main()



from setuptools import setup

setup(
    name='tiros',
    version='1.0.17',
    packages=['tiros'],
    install_requires=['requests>=2.11.1', 'boto3>=1.3.1', 'datetime>=4.1.1'],
    entry_points={
        'console_scripts': [
            'tiros = tiros.cli:main',
        ],
    },
)
